# a snippet to be used with pyrasite shell to inspect hanging threads - stack traces
# from each thread

import sys, traceback

def show_thread_stacks():
    for thread_id, frame in sys._current_frames().items():
        print('\n--- Stack for thread {t} ---'.format(t=thread_id))
        traceback.print_stack(frame, file=sys.stdout)

show_thread_stacks()



__version__ = "0.12.4"

from .events import EvConfUpdate, EvExternal, EvStart, EvStop, EvTick, attach_handlers
from .Gateway import Gateway
from .modules.Module import ConfigScopeEnum, Module
from .UuAppClient import UuAppClient, UuCmdSession



import json
import os
from contextlib import contextmanager

import pytest
from uun_iot import Gateway
from uun_iot.UuAppClient import UuAppClient

# set correct server credentials in tests/config.json and set this to True to enable server testing
TEST_SERVER_COMM = False


# @pytest.fixture(scope="class")
def _config_location(tmpdir_factory, config_object):
    path = str(tmpdir_factory.mktemp("data").join("config.json"))
    with open(path, "w") as f:
        json.dump(config_object, f)

    return path


# @pytest.fixture(scope="class")
def _config_object():
    """Overwrite with custom configuration."""
    #return json.load(open("tests/config.json"))
    return {}


# @pytest.fixture(scope="class")
def _config_gateway(config_object):
    """Return a gateway subkey of configuration."""
    return config_object["gateway"]


# @pytest.fixture(scope="class")
def _init_modules():
    """Overwrite with custom init function."""
    return None

def _init_events():
    """Overwrite with custom event init function."""
    return None


# @pytest.fixture(scope="class")
def _gateway(config_location, init_modules, init_events):
    """Supply a config_object fixture in given module to overwite default file configuration."""

    if init_events:
        yield Gateway(config_location, None, init_events)
    else:
        yield Gateway(config_location, init_modules)

    try:
        os.remove("oidc-token")  # delete cached token
    except FileNotFoundError:
        pass

    try:
        os.remove("dummy")  # delete created files
    except FileNotFoundError:
        pass


# @pytest.fixture(scope="class")
def _started_gateway(gateway):
    with gateway as g:
        yield g


# @pytest.fixture(scope="class")
def _fileconfig(config_location):
    # return current configuration
    with open(config_location) as f:
        config = json.load(f)
    return config


# @pytest.fixture(scope="class")
def _uuappclient(fileconfig):
    return UuAppClient(fileconfig)


##### FIXTURES ######


@pytest.fixture(scope="class")
def config_location(tmpdir_factory, config_object):
    return _config_location(tmpdir_factory, config_object)


@pytest.fixture(scope="class")
def config_location_fn(tmpdir_factory_fn, config_object_fn):
    return _config_location(tmpdir_factory_fn, config_object_fn)


@pytest.fixture(scope="class")
def config_object():
    return _config_object()


@pytest.fixture(scope="function")
def config_object_fn():
    return _config_object()


@pytest.fixture(scope="class")
def config_gateway(config_object):
    return _config_gateway(config_object)


@pytest.fixture(scope="function")
def config_gateway_fn(config_object_fn):
    return _config_gateway(config_object_fn)


@pytest.fixture(scope="class")
def init_modules():
    return _init_modules()


@pytest.fixture(scope="function")
def init_modules_fn():
    return _init_modules()

@pytest.fixture(scope="class")
def init_events():
    return _init_events()


@pytest.fixture(scope="function")
def init_events_fn():
    return _init_events()


@pytest.fixture(scope="class")
def gateway(config_location, init_modules, init_events):
    yield from _gateway(config_location, init_modules, init_events)


@pytest.fixture(scope="function")
def gateway_fn(config_location_fn, init_modules_fn):
    yield from _gateway(config_location_fn, init_modules_fn, init_events_fn)


@pytest.fixture(scope="class")
def started_gateway(gateway):
    yield from _started_gateway(gateway)

@pytest.fixture(scope="function")
def started_gateway_fn(gateway_fn):
    yield from _started_gateway(gateway_fn)


@pytest.fixture(scope="class")
def fileconfig(config_location):
    return _fileconfig(config_location)


@pytest.fixture(scope="function")
def fileconfig_fn(config_location_fn):
    return _fileconfig(config_location_fn)


@pytest.fixture(scope="class")
def uuappclient(fileconfig):
    return _uuappclient(fileconfig)


@pytest.fixture(scope="function")
def uuappclient_fn(fileconfig_fn):
    return uuappclient(fileconfig_fn)



import logging
from datetime import datetime
from typing import Callable, Dict, Optional

import requests
from uun_iot import EvTick
from uun_iot.utils import get_iso_timestamp

logger = logging.getLogger(__name__)


class Heartbeat:
    online: bool

    """
    Ping server in short periods to let server know if the gateway is online and
    send a little info about the gateway.

    Args:
        uucmd: function ``(dto_in) -> requests.Response``, the function takes
            an argument with data to be sent to the heartbeat uuCmd endpoint. It
            returns the reponse formed using :class:`requests.Response`.
    """

    id = "heartbeat"

    def __init__(self, uucmd: Callable[[Dict], Optional[requests.Response]]):
        self.online = False
        self._uucmd = uucmd

    @EvTick.subscribe
    def beat(self):
        """Determine online status and send a little information about gateway to uuApp.

        The online status is determined using ``self._uucmd`` uuCmd. The information sent includes

            - boot timestamp (in ISO timestamp)
            - CPU usage in percent
            - RAM usage in percent

        Executed on each timer tick.
        """

        try:
            # in docker, psutil is not installed
            import psutil

            boot_timestamp = get_iso_timestamp(
                datetime.fromtimestamp(psutil.boot_time())
            )
            cpu = psutil.cpu_percent()
            ram = psutil.virtual_memory().percent
        except ImportError:
            boot_timestamp = cpu = ram = None

        dto_in = {
            "bootTimestamp": boot_timestamp,
            "cpuUsage": cpu,
            "ramUsage": ram,
        }

        response = self._uucmd(dto_in)
        if response:
            if response.status_code >= 200 and response.status_code < 300:
                if not self.online:
                    logger.info("online")
                    self.online = True
            else:
                if self.online:
                    logger.warning("uuApp server error")
                    self.online = False
        else:
            if self.online:
                logger.warning("network offline")
                self.online = False



import logging
import os
import selectors
import socket
import threading
import typing as t

from uun_iot import EvExternal, EvStart, EvStop

from ..typing import ModuleId
from .Module import Module

logger = logging.getLogger(__name__)


class SocketCmdError(Exception):
    """General exception raised when trying to execute received command."""


class InvalidMsg(SocketCmdError):
    """Received ill-formed command."""


class CommandNotRegistered(SocketCmdError):
    """Command name was not registered using @on("external")."""


class CommandFailed(SocketCmdError):
    """While executing command handler, an exception occured."""


class SocketServer(Module):
    id: ModuleId = "socket"
    #: path to socket, a command server socket
    socket_path: str

    #: pipe used for notifying socket server to shut down
    _ping_socket_pipe: t.Tuple[int, int]

    _ev_external: EvExternal

    def __init__(self, config: dict, ev_external: EvExternal):
        super().__init__(config=config)

        self._path = None
        self._ev_external = ev_external
        try:
            self._path = self._config_manager("path")
        except KeyError as e:
            raise ValueError("No path specified.") from e

        try:
            os.unlink(self._path)
        except OSError as e:
            if os.path.exists(self._path):
                raise ValueError("Could not create socket.") from e

        self._ping_socket_pipe = os.pipe()

    @EvStart.subscribe_blocking
    def start(self, evs):
        """Start a socket command server.

        Block until the server is ready and then serve in a daemon thread.
        """
        if not self._path:
            return

        runev, _ = evs
        sel = selectors.DefaultSelector()

        pipe_r, _ = self._ping_socket_pipe
        # write to ping_socket_pipe_w to cause sel.select() to fire,
        # effectively aborting waiting for new connection (via accept())
        sel.register(
            pipe_r,
            selectors.EVENT_READ,
            lambda pipe, mask: logger.debug("stopping socket server"),
        )

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(self._path)
        os.chmod(self._path, 0o600)
        sock.listen(1)
        sock.setblocking(False)
        sel.register(sock, selectors.EVENT_READ, self._accept_connection)
        logger.debug("Opening socket server at '%s'", self._path)

        def serve():
            with sock:
                while runev.is_set():
                    logger.debug("Socket server: waiting...")
                    events = sel.select()
                    logger.debug("received new socket event")
                    for key, mask in events:
                        callback = key.data
                        callback(key.fileobj, mask)

        threading.Thread(target=serve, daemon=True).start()

    @EvStop.subscribe
    def stop(self):
        """Stop socket server. Clear runev first."""
        if not self._path:
            return
        _, pipe_w = self._ping_socket_pipe
        os.write(pipe_w, b"!")

    def _parse_cmd(self, msg: str) -> t.Optional[str]:
        """Parse and execute external command.

        Args:
            msg:  in format 'cmd_name[<space>arguments]', '[...]' being optional

        Raises:
            ValueError: msg is an invalid command
        """
        try:
            cmd, arg = msg.split(" ", maxsplit=1)
        except ValueError as e:
            cmd = msg
            arg = None
            # raise InvalidMsg(
            #    "A space delimiter must be present to destinguish command name"
            #    " and its arguments."
            # ) from e

        callback = self._ev_external.get_handler(cmd)
        if callback is None:
            raise CommandNotRegistered(f"Command '{cmd}' was not registered.")

        # currently, only a single module command origin is supported
        # for _, callback in self._handlers[cmd].items():
        try:
            out = callback((cmd, arg))
        except Exception as e:
            logger.error(
                "Exception occured in @external.%s for message '%s': %s",
                cmd,
                msg,
                repr(e),
            )
            raise CommandFailed(repr(e)) from e

        # if isinstance(out, str):
        #    out = out.encode("utf-8")
        logger.debug("External command '%s' processed.", cmd)
        return out

    def _accept_connection(self, sock: socket.socket, mask):
        conn, _ = sock.accept()  # sock is ready to accept, this will not block
        logger.debug("A client connected to socket.")
        with conn:  # automatically closes conn
            # expected time [s] to write to socket by client, close it after timeout
            conn.settimeout(1)
            with conn.makefile(mode="rw", encoding="utf-8") as conn_file:
                try:
                    # read a single line of input
                    line = conn_file.readline()
                    logger.debug("Received '%s'", line.strip())
                    response = ""
                    try:
                        cmd_output = self._parse_cmd(line.strip("\n\r"))
                        if cmd_output is None:
                            response = "0"
                        else:
                            response = "0 " + cmd_output
                    except InvalidMsg:
                        response = "1 Invalid message!"
                    except CommandNotRegistered:
                        response = "1 Unknown command!"
                    except CommandFailed as e:
                        response = f"1 Command failed! Exception: {str(e)}"

                    try:
                        conn_file.write(response + "\n")
                        conn_file.flush()
                        logger.debug("Written: '%s'", response)
                    except BrokenPipeError:
                        logger.info(
                            "Error while writing to socket. The receiver maybe was"
                            " not waiting for response?"
                        )

                except TimeoutError:
                    logger.debug(
                        "Socket connection from client accepted but no command was"
                        " written or could not deliver command output. Did you"
                        " forget to end the command with a newline?"
                    )



"""Initialize modules."""

import json
import logging
import typing as t

from uun_iot.UuAppClient import UuAppClient

from ..events import EvConfUpdate, Event, EvExternal, EvTick, attach_handlers, merge_events
from ..utils import maybe_get_instance
# from .BaseHealthCheck import register_basehealthcheck
from .ConfigUpdater import ConfigUpdater
from .Heartbeat import Heartbeat
from .SocketServer import SocketServer

logger = logging.getLogger(__name__)
loggerc = logging.getLogger(__name__ + "ConfigUpdaterUucmd")


def init_library_module_events(
    config_path: t.Optional[str],
    config: t.Dict,
    uuclient: UuAppClient,
    events: t.Optional[t.Iterable[Event]] = None,
) -> t.Iterable[Event]:
    full_config = config
    config = config["gateway"] if "gateway" in config else config
    modules = []

    ev_confupdate = maybe_get_instance(EvConfUpdate, events) or EvConfUpdate(full_config)
    ev_tick = maybe_get_instance(EvTick, events) or EvTick(full_config)
    ev_external = maybe_get_instance(EvExternal, events) or EvExternal(full_config)
    evs = [ev_confupdate, ev_tick, ev_external]

    ### ConfigUpdater
    endpoint_uucmd_config = (
        config.get("uuApp", {})
        .get("uuCmdList", {})
        .get("gatewayGetConfiguration", None)
    )
    if endpoint_uucmd_config:

        def uucmd_config_update():
            # create uucmd function, separate from other cmd due to nature of Config
            r, exc = uuclient.get(endpoint_uucmd_config, log_level=logging.DEBUG)
            if r is None or exc is not None:
                return None

            try:
                new_config = r.json()
                # this is not an actual configuration key, only a remnant of server communication
                if "uuAppErrorMap" in new_config:
                    del new_config["uuAppErrorMap"]
                if not new_config:
                    loggerc.debug(
                        "Received empty JSON configuration from server, ignoring."
                    )
                    return None
            except json.decoder.JSONDecodeError:
                loggerc.warning("Received invalid configuration JSON from server")
                loggerc.debug("Invalid response: %s", r.text)
                return None

            return new_config

        def save_config_to_file(config):  # -> bool:
            if config_path is None:
                return  # False
            try:
                with open(config_path, "w") as f:
                    f.write(json.dumps(config, indent=4))
                    return  # True
            except:
                return  # False

        def update_callback(new_config):
            ev_tick.update(new_config)
            ev_confupdate.fire_handlers(new_config, block=False)

        modules.append(
            ConfigUpdater(
                full_config, uucmd_config_update, save_config_to_file, update_callback
            )
        )

    ### SocketServer
    try:
        modules.append(SocketServer(config, ev_external))
        logger.info("socket server successfuly initialized")
    except ValueError as er:
        logger.info("socket server was not initialized: %s", str(er))

    ### Heartbeat
    endpoint_uucmd_heartbeat = (
        config.get("uuApp", {}).get("uuCmdList", {}).get("gatewayHeartbeat", None)
    )
    if endpoint_uucmd_heartbeat:

        def uucmd_heartbeat(dto_in):
            resp, exc = uuclient.post(
                endpoint_uucmd_heartbeat, dto_in, log_level=logging.DEBUG
            )
            if exc is None:
                return resp
            return None

        modules.append(Heartbeat(uucmd_heartbeat))

    all_evs = attach_handlers(full_config, modules, evs)
    # all_evs does not containg EvExternal as none of the base modules directly
    # subscribe to it - attach_handlers does not find it in the passed modules.
    # Instead, the SocketServer indirectly uses it so we need to
    # forcefully include it in returned events so that user modules subscribing to
    # EvExternal get the same instance of the event which was registered with SocketServer
    return merge_events(all_evs, evs)



import logging
import typing as t
from dataclasses import dataclass

from ..exceptions import EventCannotRegisterHandler
from ..utils import classproperty
from .Event import Event, EventType, Register

logger = logging.getLogger(__name__)

SelfM = t.Any
F = t.TypeVar("F")


class DuplicateKeyError(ValueError):
    pass


K = t.TypeVar("K")
V = t.TypeVar("V")


@dataclass
class DuplicateKeyErrStruct(t.Generic[K, V]):
    key: K
    current: V
    new: V


def dict_throw_on_duplicate_keys(
    it: t.Iterable[t.Tuple[K, V]],
) -> t.Dict[K, V]:
    d = {}
    for k, v in it:
        if k in d:
            raise DuplicateKeyError(DuplicateKeyErrStruct(k, d[k], v))
        d[k] = v
    return d


def dict_append_on_duplicate_keys(
    it: t.Iterable[t.Tuple[K, V]],
) -> t.Dict[K, t.List[V]]:
    d = {}
    for k, v in it:
        if k not in d:
            d[k] = [v]
        else:
            d[k].append(v)
    return d


def populate_handlers(
    it: t.Iterable[t.Tuple[K, V]], fn: t.Callable[[DuplicateKeyErrStruct], str]
) -> t.Dict[K, V]:
    try:
        return dict_throw_on_duplicate_keys(it)
    except DuplicateKeyError as e:
        einfo = e.args[0]
        raise ValueError(fn(einfo)) from e


class EvExternal(Event):
    """
    Event for mediating socket commands. This class acts as an intermediary between the
    socket server and handlers. See :class:`uun_iot.modules.ConfUpdater` for more
    details.

    Examples:

        - configuration:

            .. code-block:: json

                {
                    "gateway": {
                        "socket": {
                            "path-to-socket-file.socket"
                        }
                    }
                }

        .. code-block:: python

            from uun_iot import EvExternal

            class ExternallyControlledModule:
                @EvExternal.subscribe_command("action1")
                def handle_cmd(self, msg):
                    cmd, msg = msg
                    assert(cmd == "action1")
                    print(msg)

            >>> echo "action1 This message was created outside of the main Python app." | nc -U path/to/unix/socket.sock
            >>> # handle_cmd is called and prints
            >>> 'This message was created outside of the main Python app.'
    """

    event_type = EventType.EXTERNAL
    _handlers: t.Dict[str, t.Callable]

    def __init__(self, config=None):
        self._handlers = {}
        super().__init__(config=None)

    @classmethod
    def subscribe_command(cls, action_id: str):
        """Subscribe to socket command.

        The handler is passed positional argument ``msg`` as a tuple ``(cmd, rest_msg)``,
        where cmd is the issued command and ``rest_msg`` is the rest of the
        message received by socket IPC

        Args:
            action_id: string identifier of action
        """
        if not isinstance(action_id, str):
            raise EventCannotRegisterHandler("Only string action IDs are supported.")
        return cls._register(
            Register[
                t.Callable[[SelfM, t.Tuple[str, t.Optional[str]]], t.Optional[str]]
            ],
            action_id,
        )

    def add_handlers(self, handler_list):
        # self._handlers = {info.args[0]: fn for fn, info in handler_list}
        self._handlers = populate_handlers(
            ((info.args[0], fn) for fn, info in handler_list),
            lambda einfo: f"Encountered duplicate entry for @EvExternal.handle_action, id '{einfo.key}'",
        )

    def get_handler(
        self, action_id: str
    ) -> t.Optional[t.Callable[[t.Tuple[str, t.Optional[str]]], t.Optional[str]]]:
        """Get handler with action_id, if exists.

        Args:
            action_id: identifier of handler
        """
        return self._handlers.get(action_id)

    def start(self):
        """Noop"""
        logger.debug("EvExternal has these handlers registered: %s", repr(self._handlers))

    def stop(self):
        """Noop"""
        pass



import logging
import threading
import typing as t

from ..utils import classproperty
from .Event import Event, EventType, Register

logger = logging.getLogger(__name__)

SelfM = t.Any


class EvStop(Event):
    """An Event which invokes handlers on app stop."""

    event_type = EventType.STOP
    _handlers: list

    def __init__(self, config=None):
        self._handlers = []
        super().__init__(config=None)

    @classproperty
    def subscribe(cls):
        """Subscribe to Stop event.

        The handler is run on `.stop` and it blocks until the handler returns.
        """
        return cls._register(Register[t.Callable[[SelfM], None]])

    def add_handlers(self, handler_list):
        self._handlers = [fn for fn, info in handler_list]

    def start(self):
        """No-op."""
        logger.debug("EvStop has these handlers defined: %s", repr(self._handlers))
        pass

    def stop(self):
        """Invoke handlers and block till they finish.

        Each handler is started in a separate thread, then threads are
        collected.
        """
        if not self._handlers:
            return

        threads = []
        for fn in self._handlers:
            logger.debug("Calling EvStop handler, function `%s`", fn)
            th = threading.Thread(target=fn)
            th.start()
            threads.append(th)

        for th in threads:
            logger.debug("joining EvStop thread %s", th)
            th.join()
            logger.debug("joined EvStop thread %s", th)



""" Communication with the .Main uuSubApp. """

import datetime
import json
import logging
import time
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder
from requests_toolbelt.utils import user_agent as ua

import uun_iot

from .exceptions import TokenCommandError, TokenError, UuAppClientException

logger = logging.getLogger(__name__)


class UuAppClient:
    """Library functions for communication and authentication with uuApp.

    The class gets a dictionary ``config`` containing the configuration of the application.
    The (root) keys and subkeys of interest to this class are

        - ``uuApp``: information about server uuApp

            - ``gateway``: domain gateway of the corresponding uuApp
            - ``uuAppName``: full name of the uuApp
            - ``awid``
            - ``uuCmdList``: containing uuCmd endpoint. The keys are not used
              by :class:`UuAppClient` class - they are used by the application
              user modules (most commonly specified in Python package path
              ``<package_name>/modules/__init__.py``.

        - ``uuThing``: authentication information about the IoT gateway

            - ``uuIdentity``
            - ``accessCode1``
            - ``accessCode2``

        - ``oidcGrantToken``: information about the authentication token
          service used in server communication

            - ``gateway``: (usually ``uuidentity.plus4u.net``) domain gateway of the `OIDC` service
            - ``uuAppName``: (usually ``uu-oidc-maing02``) full name of the `OIDC` uuApp
            - ``awid``
            - ``clientId``
            - ``clientSecret``
            - ``uuCmd``: (usually ``oidc/grantToken``) uuCmd enpoint for token granting
            - ``tokenPath``: (usually ``./oidc-token``) filesystem location (on
              the IoT gateway) where the token will be saved

    Args:
        config: dictionary with configuration, see above
        fn_post_request: optional function to get a token response from server,
            defaults to :meth:`requests.post`
        refresh_token_on_init: call :meth:`.refresh_token` on initialization of this class
        token_leeway: leeway (drift) between Client and Server Clock [s], defaults to 60 s
            fresh token is requested when token is to be expired in less than token_leeway [s]
            the lower the leeway, the more often is token going to be requested
            the token validity time is controlled by the server, this option accounts only for drift
    """

    _config: dict
    _token: Optional[str]
    _user_agent: str
    _token_expires_at: Optional[float]
    _auth_active: bool
    _token_leeway: float

    def __init__(
        self,
        config: dict,
        fn_token_request: Callable[..., requests.Response] = requests.post,
        refresh_token_on_init: bool = True,
        token_leeway: float = 60,
    ):
        self._config = config
        self._token = None
        self._token_expires_at = None
        self._user_agent = ua.UserAgentBuilder("uun-iot", uun_iot.__version__).build()
        self._token_leeway = token_leeway
        self._auth_active = (
            "uuApp" in self._config
            and "uuThing" in self._config
            and "oidcGrantToken" in self._config
        )
        self._fn_token_request = fn_token_request

        if refresh_token_on_init and self._auth_active:
            self.refresh_token()
        else:
            if refresh_token_on_init:
                logger.info(
                    "UuApp connection & gateway authentication is not "
                    "defined in the configuration file. Are you sure, you do not "
                    "want to use the connection? It will not be possible to "
                    "issue requests to uu servers."
                )
            logger.debug("Not granting any auth token.")

    def _store_token(self, token_json: dict) -> None:
        """Save serialized token to file."""
        with open(self._config["oidcGrantToken"]["tokenPath"], "w") as file:
            file.write(json.dumps(token_json))

    def is_token_expired(self) -> bool:
        """
        # view with respect to server-centered time
        #       <-> leeway
        # ++++++==|-----
        #         |
        #    server_token_expire
        #
        # +: when (local) now is there, token is valid
        # =: when (local) now is there, token is valid but apply for a new token
        #      because of the possible leeway between client and server clocks
        # -: when (local) now is there, token is invalid
        """
        return (
            self._token_expires_at is None
            or self._token_expires_at - self._token_leeway < time.time()
        )

    def refresh_token(self) -> bool:
        """Refresh token if expired. If authentication is disabled, noops and returns False.

        Returns False if token refresh was not needed - that is, if token is
        already loaded and valid. Refresh token and return True otherwise.

        Raises:
            TokenError: when a valid token could not be obtained due to network or server error
            TokenCommandError (subclass of TokenError): server-side uuApp returned error
        """
        if not self._auth_active:
            return False

        if self._token is None:
            # load from file
            token_path = self._config["oidcGrantToken"]["tokenPath"]
            try:
                with open(token_path) as json_file:
                    token = json.load(json_file)
                    self._token = token["id_token"]
                    self._token_expires_at = float(token["expires_at"])

                logger.debug(
                    "loaded token from %s, expires at [%s]",
                    token_path,
                    datetime.datetime.fromtimestamp(self._token_expires_at),
                )
            except IOError:
                logger.debug("loading token from %s failed, file not found", token_path)
            except json.JSONDecodeError:
                logger.debug("Invalid JSON token file.")
            except KeyError:
                logger.debug("Invalid structure of token file.")

        if self.is_token_expired():
            self._token_expires_at = self._grant_token(self._fn_token_request)
            logger.debug(
                "granted new token, expires at [%s]",
                datetime.datetime.fromtimestamp(self._token_expires_at),
            )
            return True

        return False

    def _grant_token(self, fn_post_request: Callable[..., requests.Response]) -> float:
        """Grant token.

        Refresh and save the new token.

        Args:
            fn_post_request: function to get a token response from server

        Returns:
             expiry date as unix timestamp

        Raises:
            TokenError: when a valid token could not be obtained due to network or server error
            TokenCommandError: server-side uuApp returned error
        """
        scope_host = self._config["uuApp"]["gateway"]
        scope_context = self._config["uuApp"]["uuAppName"]
        scope_awid = self._config["uuApp"]["awid"]
        post_data = {
            "grant_type": "password",
            "username": None,
            "password": None,
            "accessCode1": self._config["uuThing"]["accessCode1"],
            "accessCode2": self._config["uuThing"]["accessCode2"],
            "scope": f"openid https://{scope_host}/{scope_context}/{scope_awid}",
        }

        try:
            post_data.update(
                {
                    "client_id": self._config["oidcGrantToken"]["clientId"],
                    "client_secret": self._config["oidcGrantToken"]["clientSecret"],
                }
            )
        except KeyError:
            # does not matter if these are not present - not required - depends on server settings
            pass

        headers = {"Content-Type": "application/json", "User-Agent": self._user_agent}
        uucmd = self._config["oidcGrantToken"]["uuCmd"]

        host = self._config["oidcGrantToken"]["gateway"]
        uu_app_name = self._config["oidcGrantToken"]["uuAppName"]
        awid = self._config["oidcGrantToken"]["awid"]
        url = f"https://{host}/{uu_app_name}/{awid}/{uucmd}"
        response = None
        try:
            response = fn_post_request(url, headers=headers, json=post_data, timeout=20)
            response.raise_for_status()
        except requests.HTTPError as e:
            assert response is not None
            logger.log(
                logging.DEBUG,
                "`%s` POST response (%s): %s",
                uucmd,
                response.status_code,
                response.text,
            )
            status = response.status_code
            try:
                error_json = response.json()
            except json.JSONDecodeError as exc:
                raise TokenError(
                    "A server error occured while getting a token and could not decode"
                    " JSON error object."
                ) from exc
            if "uuAppErrorMap" in error_json:
                raise TokenCommandError(error_json["uuAppErrorMap"], status) from e
            raise TokenError(
                "Server error occured but could not get more information."
            ) from e
        except requests.RequestException as e:
            logger.log(logging.WARNING, "`%s` POST RequestException when getting a new token: %s", uucmd, str(e))
            raise TokenError(
                "A network error occured when requesting new token."
            ) from e

        try:
            token_json = response.json()
            self._token = token_json["id_token"]
            token_json["expires_at"] = time.time() + float(token_json["expires_in"])
        except json.JSONDecodeError as e:
            raise TokenError(
                "While getting a token, could not decode received JSON server response."
            ) from e
        except KeyError as e:
            raise TokenError("Token did not have required fields.") from e
        except ValueError as e:
            raise TokenError("Token has invalid value types.") from e

        self._store_token(token_json)
        return token_json["expires_at"]

    def get_uucmd_url(self, uucmd: str) -> str:
        """Return fully quallified URL from relative uucmd string.

        Args:
            uucmd: uucmd in relative path, eg. ``gateway/getWeather``
        """
        host = self._config["uuApp"]["gateway"]
        uu_app_name = self._config["uuApp"]["uuAppName"]
        awid = self._config["uuApp"]["awid"]
        return f"https://{host}/{uu_app_name}/{awid}/{uucmd}"

    def get_auth_headers(self, content_type: str = "application/json"):
        """Get request headers with Bearer authentication.
        The token is refreshed automatically, if expired.

        Headers also includes specified ``Content-type`` and uun-iot's ``User-Agent``.
        """
        if self.is_token_expired():
            self.refresh_token()
        return {
            "Authorization": f"Bearer {self._token}",
            "Content-type": content_type,
            "User-Agent": self._user_agent,
        }

    def custom_request(
        self,
        custom_send_f: Callable[[object], requests.Response],
        dto_in: Optional[object] = None,
        label: Optional[str] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        Custom request wrapper for catching :mod:`requests`-related and token-acquiring-related
        Exceptions. Token errors, request exceptions, including HTTP non-success status codes,
        are suppressed, logged with specified level and returned.

        Supply your own :attr:`custom_send_f` function (taking :attr:`dto_in` as
        only argument) and outputting a :attr:`requests.Response`. To make an
        authenticated request, you should use authentication headers available
        through :meth:`.get_auth_headers`. Note that you can specify the headers
        either directly during *definition*, or during *evaluation* of
        :attr:`custom_send_f`. The latter is preferred due to handling TokenErrors
        in a unified way with other exceptions. Note that token acquisition is not in
        scope of :meth:`.custom_request` and has to be supplied by caller.

        See also :meth:`.get_uucmd_url` and :meth:`.get_auth_headers` for
        constructing own function.

        .. note::
            Example for ``custom_send_f`` follows.

            .. code-block:: python

                uucmd = "gateway/heartbeat"
                full_url = uuappclient.get_uucmd_url(uucmd)
                def handler(dto_in):
                    return requests.get(full_url, headers=self.get_auth_headers(), json=dto_in_, timeout=20)
                uuappclient.custom_request(handler, {"key": "value"}, uucmd)


        This function will
            - call :attr:`custom_send_f` with :attr:`dto_in` positional
              argument,
            - raise for non-success HTTP status codes
            - catch :attr:`requests.HTTPError` and
              :attr:`requests.RequestException`,
            - log the exceptions and
            - returns response along with the exception

        Args:
            custom_send_f: custom send function which is used instead of
                :meth:`requests.get` label: label used in log messages to identify the request
            dto_in: dictionary data to send. If ``None``, pass ``{}`` to :attr:`custom_send_f`.
            log_level: optional logging level for possible exceptions, defaults to ``logging.WARNING``

        Returns: Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not
                  200 HTTP status code) exception was raised
                - ``exc``: :class:`requests.RequestException` or :class:`TokenError` if the same
                  exception was raised during request. ``None`` if no exception during request was raised

        Raises: None
        """

        if dto_in is None:
            dto_in = {}

        response = None
        exc = None
        try:
            response = custom_send_f(dto_in)
            response.raise_for_status()
        except requests.HTTPError as e:
            assert response is not None
            logger.log(
                log_level,
                "`%s` response (%s): %s",
                label,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(log_level, "`%s` RequestException: %s", label, str(e))
            exc = e
        except TokenError as e:
            logger.log(log_level, "`%s` TokenError: %s", label, str(e))
            exc = e

        return response, exc

    def get(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        Get request using an authenticated Bearer token.
        Token errors, request exceptions, including HTTP non-success status codes, are suppressed,
        logged with specified level and returned.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: UuCmd in relative path format, eg. ``gateway/heartbeat``
            dto_in: data to send
            log_level: optional logging level for possible exceptions,
                defaults to ``logging.WARNING``

        Returns:
            Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not 200 HTTP status code)
                  exception was raised
                - :class:`requests.RequestException` exc is None if no exception
                  during request was raised
        """

        full_url = self.get_uucmd_url(uucmd)

        def get_f(dto_in_):
            # ask for headers in the handler function, custom_request first refreshes token
            logger.debug("`%s` GET request, payload: %s", uucmd, dto_in)
            return requests.get(
                full_url, headers=self.get_auth_headers(), json=dto_in_, timeout=20
            )

        return self.custom_request(get_f, dto_in, uucmd, log_level)

    def get_ignore_http_err(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> requests.Response:
        """
        GET request with authentication. Ignore HTTP errors, raise errors for other network errors.
        Logs all errors (including HTTP non-200) with :attr:`log_level` severity.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            log_level: logging library level of network error, default logging.WARNING

        Raises: :class:`requests.RequestException` or :class:`TokenError`
        """

        r, exc = self.get(uucmd, dto_in, log_level)
        if exc is not None:
            try:
                raise exc
            except requests.HTTPError:
                pass
        assert r is not None
        return r

    def post(
        self,
        uucmd: str,
        dto_in: Optional[Union[dict, MultipartEncoder]] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        POST request using an authenticated Bearer token.
        Request exceptions, including HTTP non-success status codes, are suppressed, logged with specified level and returned.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: UuCmd in relative path format, eg. ``gateway/heartbeat``
            dto_in: data to send, dictionary or MultipartEncoder data
            log_level: optional logging level for possible exceptions,
                defaults to ``logging.WARNING``

        Returns:
            Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not 200 HTTP status code)
                  exception was raised
                - :class:`requests.RequestException` exc is None if no exception
                  during request was raised

        Raises: ValueError if ``dto_in`` type is not ``dict`` or ``MultipartEncoder``
        """

        if dto_in is None:
            dto_in = {}

        full_url = self.get_uucmd_url(uucmd)
        if isinstance(dto_in, dict):
            headers = self.get_auth_headers
            def post_f(dto_in_):
                logger.debug("`%s` POST request, payload: %s", uucmd, dto_in)
                return requests.post(full_url, headers=headers(), json=dto_in_, timeout=20)
        elif isinstance(dto_in, MultipartEncoder):
            headers = partial(self.get_auth_headers, dto_in.content_type)
            def post_f(dto_in_):
                logger.debug("`%s` MULTIPART request, payload: %s", uucmd, dto_in)
                return requests.post(full_url, headers=headers(), data=dto_in_, timeout=20)
        else:
            raise ValueError(f"Invalid dto_in type: {type(dto_in)}")

        return self.custom_request(post_f, dto_in, uucmd, log_level)

    def post_ignore_http_err(
        self,
        uucmd: str,
        dto_in: Optional[Union[dict, MultipartEncoder]] = None,
        log_level: int = logging.WARNING,
    ) -> requests.Response:
        """
        POST request with authentication. Ignore HTTP errors, raise errors for other network errors.
        Logs all errors (including HTTP non-200) with :attr:`log_level` severity.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            log_level: logging library level of network error, default logging.WARNING

        Raises: :class:`requests.RequestException` or :class:`TokenError`
        """
        r, exc = self.post(uucmd, dto_in, log_level)
        if exc is not None:
            try:
                raise exc
            except requests.HTTPError:
                pass
        assert r is not None
        return r

    def multipart(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """POST request with authentication and MULTIPART encoded data
        with oidc2 authentication. Useful for sending binary data (images, ...). See
        https://toolbelt.readthedocs.io/en/latest/user.html#multipart-form-data-encoder
        for information about multipart encoder.

        Pass data to dto_in, they will be transformed using MultipartEncoder and passed to
        :meth:`UuAppClient.post` (see for more information and usage).
        """

        if dto_in is None:
            return self.post(uucmd, None, log_level)
        multipart_encoded = MultipartEncoder(fields=dto_in)
        return self.post(uucmd, multipart_encoded, log_level)


class UuCmdSession:
    """
    Send all data with UuCmd in one session to avoid creating multiple connections.

    Args:
        uuclient: initialized authentication client
        uucmd: UuCmd in relative path format, eg. ``gateway/getWeather``
        log_level: ``logging`` library log level to use for network and HTTP errors,
            defaults to :attr:`logging.WARNING`
    """

    def __init__(
        self, uuclient: UuAppClient, uucmd: str, log_level: int = logging.WARNING
    ):
        self._uucmd = uucmd
        self._loglevel = log_level
        self._url = uuclient.get_uucmd_url(uucmd)
        self._session = requests.Session()
        self._uuclient = uuclient

    def update_headers(self):
        """Update session auth headers ensured to containing a valid Bearer token."""
        self._session.headers.update(self._uuclient.get_auth_headers())

    def get(
        self, data: Optional[dict] = None, log_level: int = logging.WARNING
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """Authenticated GET request with session.
        See :meth:`UuAppClient.get` for more information."""

        full_url = self._uuclient.get_uucmd_url(self._uucmd)

        def get_f(dto_in_):
            self.update_headers()
            logger.debug("`%s` GET request, payload: %s", self._uucmd, data)
            return self._session.get(full_url, json=dto_in_, timeout=20)

        return self._uuclient.custom_request(get_f, data, self._uucmd, log_level)

    def post(
        self, data: Optional[dict] = None, log_level: int = logging.WARNING
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """Authenticated POST request with session.
        See :meth:`UuAppClient.post` for more information."""

        full_url = self._uuclient.get_uucmd_url(self._uucmd)

        def post_f(dto_in_):
            self.update_headers()
            logger.debug("`%s` POST request, payload: %s", self._uucmd, data)
            return self._session.post(full_url, json=dto_in_, timeout=20)

        return self._uuclient.custom_request(post_f, data, self._uucmd, log_level)



"""
Library events are defined in this module. Below is an overview of existing
events and general information regarding subscriptions to events.

================= =======================================================================
    event                   description
================= =======================================================================
 ``EvTick``        timer has ticked (timer configured in JSON config)
 ``EvStart``       app has started (via :meth:`~Gateway.start`)
 ``EvStop``        app is stopping, please end module's action as soon as possible
 ``EvExternal``    a registered command was issued using socket IPC
 ``EvConfUpdate``  gateway's configuration was updated
================= =======================================================================

Event subscriptions:

    Event handlers in classes (user modules) are registered by storing
    additional information together with the class's method, using
    corresponding `@<event>.subscribe`. This information includes event type
    and possibly additional information needed to identify the handler.

    This information is then collected from all passed instances using
    :func:`attach_handlers` and events with registered handlers are returned.
    There, appropriate events are created and the corresponding handlers are
    registered to corresponding events.

Advanced handler extraction:

    Alternatively, the :func:`attach_handlers` can register handlers to already
    existing events. You can even define your own events! See
    :func:`attach_handlers` for technical details.

Custom events:

    If you want to define your own events independently from this library, the
    only requirement is to subclass from the :class:`Event` class in order to
    be consistend with existing event handling logic. Have a look at existing
    events.

    When defining ways to store event information along with the handler in
    your custom event, you need to implement logic to resolve
    :attr:`HandlerInfo.event_type` to your custom event, while preserving
    existing resolver given in :func:`get_event_class`. For example, define
    :attr:`Event.event_type` to a unique (among all possible subclasses of
    Event) value and use the existing infrastructure

        .. code-block:: python

            import typing as t
            from uun_iot.events import Event, Register, get_event_class, attach_handlers
            from uun_iot.utils import classproperty

            # you can specify details to provide type-checking for handlers
            HandlerSignature = t.Callable[..., None]

            class MyEvent(Event):
                event_type = "myev"
                def __init__(self, config):
                    self._handlers = []
                    super().__init__(config)

                @classproperty
                def subscribe(cls):
                    return cls._register(Register[HandlerSignature])

                def add_handlers(self, handler_list):
                    self._handlers = [fn for fn, info in handler_list]

                def start(self):
                    pass
                def stop(self):
                    pass

            def get_extended_event_class(ev_type):
                if ev_type == "myev":
                    return MyEvent
                else:
                    return get_event_class(ev_type)

            class TestClass:
                @MyEvent.subscribe
                def method(self):
                    pass

            # then, you can attach handlers such as:
            evs = attach_handlers({}, [TestClass()], get_ev_class=get_extended_event_class)
            print(evs)

Note:

    Note about standard Python behaviour. When a class's instance is
    created, the method is bound such that its first argument is the
    instance (the `self` argument). When invoking the instance's
    method, the augmented function is called instead.

Warning:

    Registering handlers not belong to any class (ie. handlers which are not
    methods) is currently not supported. Future versions might allow for
    registering "bare" handler functions.
"""

import itertools
import typing as t

from ..utils import groupby_unsorted, maybe_get_instance
from .EvConfUpdate import EvConfUpdate
from .Event import (Event, EventType, HandlerInfo, Register,
                    extract_event_handlers_from_object)
from .EvExternal import EvExternal
from .EvStart import EvStart
from .EvStop import EvStop
from .EvTick import EvTick


def get_event_class(event_type: EventType):
    """Event factory

    Function matching event type identifier in :class:`HandlerInfo` struct to
    particular event constructor.

    Args:
        event_type: event ID (:class:`EventType`)

    Raises:
        ValueError: When the given :attr:`event_type` is not recognized
    """
    # match event_type:
    #    case EventType.START:
    #        return EvStart
    #    case EventType.STOP:
    #        return EvStop
    #    case EventType.CONF_UPDATE:
    #        return EvConfUpdate
    #    case EventType.TICK:
    #        return EvTick
    #    case EventType.EXTERNAL:
    #        return EvExternal
    #    case _:
    #        raise ValueError("Event type does not exists.")

    if event_type == EventType.START:
        return EvStart
    elif event_type == EventType.STOP:
        return EvStop
    elif event_type == EventType.CONF_UPDATE:
        return EvConfUpdate
    elif event_type == EventType.TICK:
        return EvTick
    elif event_type == EventType.EXTERNAL:
        return EvExternal
    else:
        raise ValueError("Event type does not exists.")


def merge_events(old: t.Iterable[Event], new: t.Iterable[Event]) -> t.Iterable[Event]:
    """Take events from old and new iterables and merge them so that the
    result contains at most one kind of each event, events from new taking precedence.

    iterators are consumed and new ones are returned
    """
    tmp_dict = {ev.event_type: ev for ev in old}
    tmp_dict_new = {ev.event_type: ev for ev in new}
    tmp_dict.update(tmp_dict_new)
    return tmp_dict.values()


def attach_handlers(
    config: t.Optional[t.Dict],
    object_list: t.Iterable[t.Any],
    events: t.Optional[t.Iterable[Event]] = None,
    get_ev_class: t.Optional[t.Callable[[EventType], t.Type[Event]]] = None,
) -> t.Collection[Event]:
    """Attach event handlers to :class:`Event` instances.

    Args:
        config: gateway config
        object_list: list of (instantiated) modules with handlers (defined via
            @Event decorator). Alternatively, you can pass the class itself if the
            handlers were already registered locally.
        events: optionally, pass a list of already instantiated events and
            handlers will be added to these events
        get_ev_class: if you define custom events, supply your own map of event
            ID to Event subclass (and preferably pass the event instance in
            :attr:`events`) inherited from :class:`EventType`

    Returns:
        a list where each Event instance is present
          (at most once for each subclass of Event) if it had handlers defined
    """
    if events is None:
        events = []
    if get_ev_class is None:
        get_ev_class = get_event_class

    def get_evtype(item: t.Tuple[t.Callable, HandlerInfo]) -> EventType:
        _, ev_data = item
        return ev_data.event_type

    def ev_with_handlers(config, ev_type, handlers):
        ev_subclass = get_ev_class(ev_type)

        ev = maybe_get_instance(ev_subclass, events)
        if ev is None:
            ev = ev_subclass(config)
        ev.add_handlers(handlers)
        return ev

    # use unsorted groupby as event_types might not be comparable when extended by user
    return [
        ev_with_handlers(config, ev_type, grouped_handler_data)
        for ev_type, grouped_handler_data in groupby_unsorted(
            list(
                itertools.chain.from_iterable(  # flatten
                    map(extract_event_handlers_from_object, object_list)
                )
            ),
            key=get_evtype,
        )
    ]
